; JCM-1 Monitor ROM
; Version 0.1
; Copyright: Joe Strosnider
; Code Date: 1/28/2018 - 2/4/2018
; Purpose: A simple monitor ROM we can use to input, view and execute code.
; Assembler Syntax: Retro Assembler - https://enginedesigns.net/apps/

.target "65C02", 64	; We're running a 65C02 Processor

; Zero Page Locations
	CmdBufPtr	= $00	; Means two things simultaneously.
							;   1) Points to the next open space in the command buffer.
							;   2) Is equal to the length of buffer contents.
								
	StrOutPtr	= $01	; Pointer to string location

	MonCurLo	= $03	; Low byte of monitor cursor location
	MonCurHi	= $04	; High byte of monitor cursor location
	
	ADIndex		= $05	; scratch-pad of the loop location of the Address Decoder
	
	StorCurLo	= $06	; Low byte of Store routine location
	StorCurHi	= $07	; High byte of Store routine location
	SMIndex		= $08	; byte index byte parser in Store routine
	
	NotHex		= $09   ; Return status of M_IsHex
	
; Other General Locations
	CmdBuf		= $0200

; Constants
				
; Hardware I/O Locations
	; ACIA 0
	ACIA0_Data			= $D000 
	ACIA0_Status 		= $D001 
	ACIA0_Command 		= $D002 
	ACIA0_Control 		= $D003

/* ******************************************************** */
/* Assembler Trick to Start ROM at $0000					*/
/* You can remove the following two lines if your ROM		*/
/* programmer allows you to save the data at $E000 offset	*/
	* = $0000
	.byte $ff
/* ******************************************************** */
	
/* ******************************************************** */
/* ROM Program Starts Here									*/

* = $e000
BIOS_ColdBoot	; Sets hardware to power on defaults, ANSI Clears screen and says hello.
	JSR InitACIA0
	JSR M_PrintImm	
	.byte $1B,"[2J",$1B,"[H","JCM-1 Computer",$00

/* ******************************************************** */
M_InitMonitor	; do stuff entering the monitor for the first time.
	STZ MonCurLo	; 
	LDA #$E0		; Initialize the monitor cursor location.
	STA MonCurHi	;

	
/* ******************************************************** */	
M_CmdPrompt	; Show the command prompt.
	STZ CmdBufPtr
	JSR M_PrintImm
	.byte $0D,$0A,$0D,$0A,$00	; Advance two lines.
	JSR M_PtMonCur				; Print Monitor Cursor
	LDA #">"					; Load up a bracket..
	JSR ACIA0_SendByte			; send the command prompt		

/* ******************************************************** */
M_Input	; Get user input.

	JSR ACIA0_GetByte   	; Get a byte.
	CMP #$1B				; Was it Escape?
	BEQ M_CmdPrompt			; Yup. Bomb immediately to prompt.
	JSR ACIA0_SendByte  	; Otherwise, spit back the char so the user knows we got it.
	LDX CmdBufPtr			; Get the Current Buffer Length. Done here to save bytes later.
	CMP #$0D				; Is it Carriage Return?
	BEQ M_ParseCmd			; Yup. We have a command in the buffer. Go parse it.
	CMP #$08				; Was it Backspace?
	BEQ	M_BkSpc				; Yup. Go deal with that.
	
	;Increment the buffer.
	CPX #$FF		; Is the buffer full?
	BEQ M_CmdPrompt	; Yes. Go prompt.
	STA CmdBuf,X    ; Otherwise, add input to buffer
	INC CmdBufPtr   ; Increment the buffer pointer
	BRA M_Input		; Go get another character.

M_BkSpc
	;Decrement the buffer.
	CPX #$00			; Check to see if it's empty.
	BEQ M_CmdPrompt  	; Buffer empty. Go prompt.
	DEC CmdBufPtr   	; Otherwise, DEC buffer pointer
	LDA #" "			; Erase...
	JSR ACIA0_SendByte	; ... the ...
	LDA #$08			; ... last ..
	JSR ACIA0_SendByte	; ... character on screen. Four more bytes than PrintIMM, but MUCH faster to do.
	BRA M_Input			; Go get another character.	

M_ParseCmd
	LDA #$0A
	JSR ACIA0_SendByte 	; give us an extra line.
	CPX #$0				; Remember, we've already loaded this with CmdBufPtr...
	BEQ CmdListImm 		; Pointer = 0 = blank input. Go list memory.
	LDX #$0				; Ready to parse. Let's start at the beginning.
	PC_Ph1	LDA CmdBuf,X	; Let's get a byte.
			CMP #$20		; Is it a space?
			BNE PC_Ph2		; Nope, continue.
			INX				; It's a space. Increment.
			CPX CmdBufPtr	; Where are we in the command buffer?
			BEQ CmdListImm 	; Reached buffer end and found nothing but spaces. Go list memory.
			BRA PC_Ph1		; Otherwise, go get another byte.
		
	PC_Ph2	AND #$DF		; Convert to uppercase.
			CMP #"G"		; [G]o?
			BEQ CmdGo  		; Yep, go execute the memory location.
			CMP #"L"		; [L]ist memory?
			BEQ	CmdList		; Yep, go list memory location
			CMP #"S"		; [S]tore Memory?
			BEQ CmdStore	; Yep, go store memory
	
	PC_Err	JSR M_PrintImm		; fell through, must be an error
			.byte "Err",$0D,$0A,$00
			BRA M_CmdPrompt

		
	CmdList	LDY CmdBufPtr		; Grab the pointer.
			DEY					; Decrement so we're looking at the last actual byte in the buffer.
			LDA CmdBuf,Y		; Grab the byte.
			AND #$DF			; Convert to uppercase.
			CMP #'L'			; Compare with L
			BEQ CmdListImm		; All we have in the buffer is L. Just go show memory.
			JSR M_AddrDecode	; Otherwise, we have some hex to decode. Let's do that.
								; Then come back and list memory immediately
								
	CmdListImm
			JSR M_PtMonCur	; Show the Monitor Cursor
			LDY #0			
	LMLoop	LDA (MonCurLo),Y	;Grab a byte.
			JSR M_PtHex			;Print it out
			LDA #" "			;Give us a little space...
			JSR ACIA0_SendByte	;... would ya, huh?
			INY					;Increment
			CPY #$10			;16 bytes?
			BNE LMLoop			;Nope, go again.
			LDA MonCurLo	;Load Monitor Cursor
			ADC #15			;Move it forward.
			STA MonCurLo	;Save it out.
			BCC LMExit		;Did it roll over?
			INC MonCurHi	;Yep. Then increment hi byte
	LMExit	JMP M_CmdPrompt

	CmdGo	JSR M_AddrDecode
			JMP (MonCurLo)
		
	CmdStore
			LDY CmdBufPtr	; Grab the pointer.
			DEY				; Decrement to look at the last char in the buffer.
			LDA CmdBuf,Y	; Grab the byte.
			AND #$DF		; Convert to uppercase.
			CMP #'S'		; Compare with S
			BEQ CSMExit		; All we have in the buffer is S. Nothing to do. Go prompt.
			INX
			LDA CmdBuf,X	; Grab the next character.
			CMP #':'		; Is it a colon?
			BEQ CSMPhaseA	; Yes. Next thing might be bytes to store. Check that.
			DEX				
			JSR M_AddrDecode ; No. Must be an address. Go decode it
			LDA CmdBuf,X	; Grab char after the address.
			CMP #':'		; Again, is it a colon?
			BNE CSMExit		; No. Error in input.
		
		CSMPhaseA	INX				; Next Char
					CPX CmdBufPtr	; have we reached buffer end?
					BEQ CSMExit		; Reached end without bytes to store.
					
					LDA MonCurLo	; Copy monitor location to store routine cursor
					STA StorCurLo	; so that store routine
					LDA MonCurHi	; can do it's
					STA StorCurHi	; magic
		
		CSMPhaseB	STZ SMIndex		; Zero the byte index.
					STZ MonCurLo    ; We need a clean byte to deal with the math.
					
		CSMPhaseC	LDA CmdBuf,X	;Get a byte
					JSR M_IsHex		;Ask if it's hex
					LDY NotHex
					BNE	CSMPhaseG	;Not Hex.
					
		CSMPhaseD	LDY SMIndex		;Which nibble are we on?
					BNE CSMPhaseE	;Second nibble.
					JSR M_ToNibble	;First nibble. Shift A into MonCurLo
					INX
					CPX CmdBufPtr	;CmdBuf End?
					BNE CSMPhaseF	;Not done. Continue
					JSR CSMStore 	;We're done.
					LDA StorCurLo	;put the monitor where it needs to be.
					STA MonCurLo	;
					LDA StorCurHi	;
					STA MonCurHi	;
					BRA CSMExit 	;Go prompt
					
		CSMPhaseE	JSR M_ToNibble	;shift A into MonCurLo	
					JSR CSMStore	;Store it
					BRA CSMPhaseH	
					
		CSMPhaseF	INC SMIndex		;Advance
					BRA CSMPhaseC	;Go again.
					
		CSMPhaseG	LDY SMIndex		;Load the Byte Index
					BEQ CSMPhaseH	;It's Zero. Need to check again.
					JSR CSMStore    ;It's one. Store previous nibble to memory.
		CSMPhaseH	LDA StorCurLo	;put the monitor where it needs to be.
					STA MonCurLo	;
					LDA StorCurHi	;
					STA MonCurHi	;
					INX
		CSMPhaseI	CPX CmdBufPtr	;Reached buffer end?
					BNE CSMPhaseB	;Nope. Keep going.
		CSMExit		JMP M_CmdPrompt	;Yup. We're done.
		
		CSMStore	LDY #0
					LDA MonCurLo		; Grab the byte to store into A
					STA (StorCurLo),Y	; store at address indicated by storcurlo
					INC StorCurLo		; Increment StorCurLo
					BNE CSMStore2		; No rollover? Then done.
					INC StorCurHi		; Rolled over. Increment StorCurHi
		CSMStore2	RTS


/* ******************************************************** */
M_AddrDecode	; decodes the hex value at the current buffer pointer and puts it in MonCurLo/Hi
; Inspired by Woz, heavily modified.

	STZ MonCurLo	; Zero out MonCur-- so nibble shift works properly.
	STZ MonCurHi	;	"
	STZ	ADIndex 	; Keeps track of times we've looped, since both index registers are taken up.
	INX				; because we're still pointing at the command char, not the value. :o)
	AD_Loop
		JSR M_IsHex 	; test for hex
		LDY NotHex
		BNE AD_Exit 	; Not hex. Bomb
								
		JSR M_ToNibble	; Is Hex. Shift it into Zero Page.
		INX             ; Advance text index.
		CPX CmdBufPtr 	; Compare X with CmdBufPtr.
		BEQ AD_Exit		; Equal to buffer end. We're done.
						; Under the buffer end? Go check to see if we're hit the 4 char limit.
	ADIndexTest
		INC ADIndex			; Advance Address Decoder index.
		LDA ADIndex
		CMP #$04			; Compare ADIndex with 4.
		BNE	AD_Loop			; if <4, we're definitely not done.
	AD_Exit
		RTS 				; We're done. Decoded address is in MonCurLo and MonCurHi

/* ******************************************************** */	
M_IsHex 	;checks that a character in A is hex.
			; if it is, returns with original value in A, and NotHex = 0
			; if it isn't, it returns with "uppercased" original value in A, and NotHex=1

	STZ NotHex	; start off assuming we will return with a good value.
	IsHex1		; determine if it's a number nibble
		LDA CmdBuf,X
		CMP #$30 	; if < 0, go check it for letter
		BCC IsHex2	
		CMP #$3A 	; if > 9, go check it for letter (or >= 9+1)
		BCS IsHex2
		RTS			; yes number. Return with value in A.

	IsHex2	;determine if it's a letter nibble.
		AND #$DF	; Convert to uppercase.
		CMP #$41 	; if < A, skip
		BCC IsHex3
		CMP #$47 	; if > F, skip (or >= F+1)
		BCS IsHex3
		ADC #$B9	; A little binary math to make our letter suitable for bit shifting
		RTS 		; yes letter. Return with value in A.

	IsHex3
		LDY #1		; not letter or number.
		STY NotHex
		RTS
		
/* ******************************************************** */		
M_ToNibble	;Move the nibble into position. Inspired by Woz Apple 1 Monitor.
		ASL		;Move digit to MSD of A
		ASL
		ASL
		ASL
		LDY #4  ; Have to shift four times.
		
	BitShift
		ASL				; Digit left, MSB to carry
		ROL MonCurLo	; Rotate into LSD
		ROL MonCurHi	; Rotate into MSD
		DEY				; Done 4 shifts?
		BNE	BitShift	; No, loop
		RTS
	
/* ******************************************************** */	
M_PrintImm	; Outputs string in memory directly after the subroutine call.
; From: http://www.6502.org/source/io/primm.htm

		PLA					; Get the low part of "return" address
		STA StrOutPtr		; Store to pointer
		PLA					; Get the high part of "return" address
		STA StrOutPtr+1		; Store to pointer
M_PI1	LDY #1				; Accounts for pointing one short. :o)
		LDA (StrOutPtr),y	; Get the next string character
		INC StrOutPtr		; update the pointer
		BNE M_PI2			; if not, we're pointing to next character
		INC StrOutPtr+1		; account for page crossing
M_PI2	ORA	#0				; Set flags according to contents of Accumulator
		BEQ M_PI3			; don't print the final NULL
		JSR ACIA0_SendByte	; write it out
		BRA M_PI1			; back around
M_PI3	INC StrOutPtr
		BNE M_PI4
		INC StrOutPtr+1		; account for page crossing
M_PI4   JMP (StrOutPtr)		; return to byte following final NULL

	
/* ******************************************************** */
M_PtMonCur		; Prints the monitor cursor on screen.
	LDA MonCurHi
	JSR M_PtHex		;send hibyte current location
	LDA MonCurLo
	JSR M_PtHex		;send lobyte current location
	LDA #":"
	JSR ACIA0_SendByte
	RTS

/* ******************************************************** */
M_PtHex	; Prints A directly to screen as a hex byte.
; Author: the one and only Woz.
; Returns with LSD in A
					
		PHA				;Save A for LSD
		LSR				;--
		LSR				; Shift MSD...
		LSR				; ... to LSD position
		LSR				;--
		JSR	PtHex		;Output hex digit
		PLA				;Restore A
PtHex	AND	#%00001111	;Mask LSD for hex print
		ORA	#"0"		;Add "0"
		CMP	#"9"+1		;Is it a decimal digit?
		BCC	ACIA0_SendByte	;Yes! output it
		ADC	#6			;Add offset for letter A-F

		;Falls through.

		
/* ******************************************************** */
ACIA0_SendByte	; Sends a byte out ACIA 0
; Byte must be in A
; Returns with registers intact			
				PHA					; Put A on the stack
	A0SB_Loop	LDA ACIA0_Status	; Get ACIA Status.
				AND #$10 			; Is TX buffer full?
				BEQ A0SB_Loop		; Yes. Go ask again.
				PLA					; No. Restore A
				STA ACIA0_Data		; And send the byte out.
				RTS

/* ******************************************************** */	
ACIA0_GetByte	; Gets a byte from ACIA 0
				; Returns with byte in A
	LDA ACIA0_Status	; Get ACIA Status.
	AND #$08 			; Was RX buffer empty?
	BEQ ACIA0_GetByte	; Yes. Go ask again.
	LDA ACIA0_Data		; No. Gimme the byte!
	RTS

/* ******************************************************** */
InitACIA0	; Set ACIA 0 for 8,N,1,115200
	LDA #%00001011		;No parity. No echo. No Interrupts
	STA ACIA0_Command
	LDA #%00010000		;115,200 Baud, 8 Data bits, 1 Stop bits
	STA ACIA0_Control
	RTS	

/* ******************************************************** */
* = $FFFA	; Vectors
.word BIOS_ColdBoot	;NMI
.word BIOS_ColdBoot	;IRQ
.word BIOS_ColdBoot	;Reset & BRK
